"""Test correct loading of the yaml file into the Parameters class."""
import subprocess
import unittest
from typing import Optional

from addict import Addict

from yamlpars import Parameters


class TestLoadingParameters(unittest.TestCase):
    """TestLoadingParameters class."""

    parameters: Optional[Parameters] = None
    files_involved = [
        "tests/test_yaml_folder/group_a.yaml",
        "tests/test_yaml_folder/group_b.yml",
    ]

    def setUp(self):
        self.parameters: Parameters = Parameters(
            yaml_folder="tests/test_yaml_folder", auto_load=False
        )
        self.assertFalse(self.parameters._successful_load)
        self.parameters.load()
        self.assertTrue(self.parameters._successful_load)
        self.assertIsNotNone(self.parameters)

    def tearDown(self):
        self.parameters = None
        self.assertIsNone(self.parameters)
        for file in self.files_involved:
            subprocess.run(f"git checkout HEAD -- {file}", shell=True, check=True)

    def test_00_load(self):
        self.assertIsNotNone(self.parameters.group_a)
        self.assertIsNotNone(self.parameters.group_b)
        self.assertRaises(AttributeError, lambda: self.parameters.group_c)

    def test_01_scalar(self):
        self.assertEqual(self.parameters.group_a.a_int, 1)
        self.assertIsInstance(self.parameters.group_a.a_int, int)
        self.assertEqual(self.parameters.group_a.a_float, 1.0)
        self.assertIsInstance(self.parameters.group_a.a_float, float)
        self.assertEqual(self.parameters.group_a.a_bool, True)
        self.assertIsInstance(self.parameters.group_a.a_bool, bool)

        self.assertEqual(self.parameters.group_b.b_int, -1)
        self.assertIsInstance(self.parameters.group_b.b_int, int)
        self.assertEqual(self.parameters.group_b.b_float, -1.0)
        self.assertIsInstance(self.parameters.group_b.b_float, float)
        self.assertEqual(self.parameters.group_b.b_bool, False)
        self.assertIsInstance(self.parameters.group_b.b_bool, bool)

    def test_02_dict(self):
        self.assertIsInstance(self.parameters.group_a.a_dictionary, dict)
        self.assertIsInstance(self.parameters.group_a.a_dictionary, Addict)
        self.assertEqual(self.parameters.group_a.a_dictionary.first_key, "first")
        self.assertEqual(self.parameters.group_a.a_dictionary.second_key, "second")
        self.assertEqual(self.parameters.group_a.a_dictionary.third_key, "third")
        self.assertEqual(
            self.parameters.group_a.a_deep_dict.with_several.layers, "works!"
        )

        self.assertIsInstance(self.parameters.group_b.b_dictionary, dict)
        self.assertIsInstance(self.parameters.group_b.b_dictionary, Addict)
        self.assertNotEqual(self.parameters.group_b.b_dictionary.first_key, "first")
        self.assertEqual(self.parameters.group_b.b_dictionary.first_key, "second")
        self.assertNotEqual(self.parameters.group_b.b_dictionary.second_key, "second")
        self.assertEqual(self.parameters.group_b.b_dictionary.second_key, "third")
        self.assertNotEqual(self.parameters.group_b.b_dictionary.third_key, "third")
        self.assertEqual(self.parameters.group_b.b_dictionary.third_key, "first")
        self.assertEqual(self.parameters.group_b.b_deep_dict.with_several.layers, {})

    def test_03_list(self):
        self.assertIsInstance(self.parameters.group_a.a_list, list)
        self.assertEqual(self.parameters.group_a.a_list, [0, 1, 2, 3])
        self.assertIsInstance(self.parameters.group_b.b_list, list)
        self.assertEqual(self.parameters.group_b.b_list, [0.0, 1.0, 2.0, 3.0])


if __name__ == "__main__":
    unittest.main()



"""TODO docstring

Matteo Pompili, 2022
"""

import pathlib
from typing import Any, Dict, List, MutableMapping, Union

from addict import Addict
from ruamel.yaml import YAML


class Parameters:
    """TODO docstring"""

    def __init__(self, yaml_folder: Union[str, pathlib.Path], auto_load: bool = True):
        super().__init__()
        self._keys: List[str] = []
        self._yaml_folder = pathlib.Path(yaml_folder)
        self._successful_load = False

        if auto_load:
            self.load()

    def load(self):
        """Load parameters into the object by parsing the yaml files.

        Raises:
            FileNotFoundError: the folder containing the yaml files does not exist.
        """
        yaml = YAML()

        if self._yaml_folder.exists():
            for suffix in ["yaml", "yml"]:
                for file in self._yaml_folder.glob(f"*.{suffix}"):
                    file_name = file.with_suffix("").name
                    with open(file, "r", encoding="UTF-8") as file_handle:
                        loaded_dict = yaml.load(file_handle)
                        if loaded_dict is not None:
                            self.__setattr__(file_name, Addict(loaded_dict))
                            self._keys.append(file_name)
        else:
            raise FileNotFoundError(f"The folder {self._yaml_folder} does not exist.")

        yaml = None
        self._successful_load = True

    @property
    def __dict__(self):
        """Return the parameters as a dictionary."""
        if self._successful_load:
            return {key: self.__getattribute__(key) for key in self._keys}
        else:
            return {}

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        if self._successful_load:
            return str(
                f'Parameters object, loading from "{self._yaml_folder}":\n'
                + self.__dict__.__repr__()
            )
        else:
            return f"Parameters object, did not load yaml folder yet."

    @staticmethod
    def _flatten_dict_gen(
        dict_to_flatten: MutableMapping[str, Any], parent_key: str, separator: str
    ):
        """https://www.freecodecamp.org/news/how-to-flatten-a-dictionary-in-python-in-4-different-ways/"""
        for k, v in dict_to_flatten.items():
            new_key = parent_key + separator + k if parent_key else k
            if isinstance(v, MutableMapping):
                yield from Parameters._flatten_dict(
                    v, new_key, separator=separator
                ).items()
            else:
                yield new_key, v

    @staticmethod
    def _flatten_dict(
        dict_to_flatten: MutableMapping, parent_key: str = "", separator: str = "."
    ) -> Dict[str, Any]:
        """https://www.freecodecamp.org/news/how-to-flatten-a-dictionary-in-python-in-4-different-ways/"""
        return dict(
            Parameters._flatten_dict_gen(dict_to_flatten, parent_key, separator)
        )

    def serialize(self) -> Dict[str, Any]:
        """Return the parameters as a dictionary of depth one.

        Example:
            >>> pars.serialize()
            { 'group_a.a_int': 1,
              'group_a.a_float': 1.0,
              'group_a.a_bool': True,
              'group_a.a_dictionary.first_key': 'first',
              'group_a.a_dictionary.second_key': 'second',
            ...
        """
        return Parameters._flatten_dict(self.__dict__)



from .parameters import Parameters
from .update import update_parameters

__title__ = "yamlpars"
__version__ = "0.2.0"
__author__ = "Matteo Pompili"
__license__ = "MIT"
__copyright__ = "Copyright 2022 Matteo Pompili"
__all__ = ["Parameters", "update_parameters"]



import pathlib
import re
from typing import Union


def lint_yaml_file(yaml_file: Union[str, pathlib.Path]) -> None:
    """Lint a yaml file, such that it passes yamllint.

    Args:
        yaml_file: the path to the file to lint.
    """

    # Fix spaces between comments
    # https://regex101.com/r/prKl93/2
    yaml_file = pathlib.Path(yaml_file)

    if not yaml_file.exists():
        raise FileNotFoundError(f"The file {yaml_file} does not exist.")

    with open(yaml_file, "r", encoding="UTF-8") as handle:
        content = handle.read()
        regex = r"^(\s*\w+[^\n\r#]+?)(?:[ \t]*)#(?:\s*)([^\n\r]*)$"
        subst = "\\g<1>  # \\g<2>"
        content_new = re.sub(regex, subst, content, 0, re.MULTILINE)
    with open(yaml_file, "w", encoding="UTF-8") as handle:
        handle.write(content_new)

    # Fix spaces at the end of the line and a final new line
    with open(yaml_file, "r", encoding="UTF-8") as handle:
        content = handle.read()
        regex = r"^(.*?)(?:\s*)$"
        subst = "\\g<1>"
        content_new = re.sub(regex, subst, content, 0, re.MULTILINE)
    with open(yaml_file, "w", encoding="UTF-8") as handle:
        handle.write(content_new + "\n")



import pathlib
from operator import getitem, setitem
from typing import Any, Dict, Optional, Union

from ruamel.yaml import YAML

from .yaml_utils import lint_yaml_file


def update_parameters(
    yaml_folder: Union[str, pathlib.Path],
    file_name: str,
    update_dict: Dict[str, Any],
    record_file_name: Optional[str] = None,
) -> None:
    """TODO docstring"""

    yaml = YAML()
    yaml.indent(mapping=2, sequence=4, offset=2)

    yaml_file = pathlib.Path(yaml_folder) / (file_name + ".yaml")
    if not yaml_file.exists():
        yaml_file = pathlib.Path(yaml_folder) / (file_name + ".yml")
        if not yaml_file.exists():
            raise FileNotFoundError(f"The file {file_name}.yaml/yml does not exist.")

    with open(yaml_file, "r", encoding="UTF-8") as file_handle:
        parameters_dict = yaml.load(file_handle)

    # For each key in the update_dict, update the corresponding key in the parameters_dict
    for key, value in update_dict.items():
        split_key = key.split(".")

        object_ref = parameters_dict

        while len(split_key) > 1:
            head = split_key.pop(0)
            object_ref = getitem(object_ref, head)
            if not isinstance(object_ref, dict):
                raise KeyError(f"Could not find {key} in {file_name}")

        # Check that the type matches
        last_value = getitem(object_ref, split_key[0])
        if not issubclass(type(last_value), type(value)):
            raise TypeError(
                f"The type of {key} in {file_name} is not the same as the type of the update value. "
                + f"It should be {type(last_value)} but is {type(value)}."
            )
        setitem(object_ref, split_key[0], value)

    # Write the updated parameters to the file
    with open(yaml_file, "w", encoding="UTF-8") as file_handle:
        yaml.dump(parameters_dict, file_handle)

    # Lint the yaml file after the changes.
    lint_yaml_file(yaml_file)

    if record_file_name is not None:
        record_file = pathlib.Path(yaml_folder) / (record_file_name + ".yaml")

        with open(record_file, "a", encoding="UTF-8") as file_handle:
            record_dict = yaml.load(file_handle)

            for key, value in update_dict.items():
                record_dict[f"record_{key}_value"] = value
                record_dict[f"record_{key}_time"] = type(value).__name__
